/*
 --  -------------------------------------------------------------------------
 --
 --      ISERVER  -  INMOS standard file server
 --
 --      udplink.c
 --
 --      Link module for SubSystems connected via the UDP-LINK series of
 --      products.
 --
 --      Copyright (c) INMOS Limited., 1989, 1990
 --
 --      17-July-2000         Kev
 --        Ported to LINUX
 --   ------------------------------------------------------------------------
*/


#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <assert.h>

#ifdef WIN_TCP
#include <time.h>
#include <winsock2.h>
#endif

#ifdef	LINUX
#include <sys/time.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#include <sys/types.h>

#include "iserver.h"

#define NULL_LINK -1
#define MAXLINE (255)
#define MAX_SEND (1400)

#ifdef WIN_TCP
#define ERRORCODE errno
#else
#define ERRORCODE errno
#endif

/**** Define local data only needed by this module ****/

PRIVATE    data_dgram	inudp;
PRIVATE          LINK	ActiveLink = NULL_LINK;
PRIVATE          BYTE	DefaultDevice[64] = "TRANSNET";
PRIVATE unsigned char	h_seq = 0;					/* host (vax) sequence no */
PRIVATE unsigned char	t_seq = 0;					/* transputer sequence no. */
PRIVATE          char *	saved_buf = NULL;
PRIVATE           int	saved_len = 0;
PRIVATE           int	Result = SUCCEEDED;         /* link error code status */

/**** Define functions only needed by this module ****/

PRIVATE           int	timed_read ( int, char *, int, long int );
PRIVATE unsigned char	prot_send ( LINK, unsigned char, unsigned char, short int, short int );
PRIVATE unsigned char	peek_send ( LINK, unsigned char, unsigned long, short int );
PRIVATE           int	data_send ( LINK, unsigned char, char *, short int );
PRIVATE           int	data_read ( LINK, char **, short int, short int, short int );


#ifdef	WIN_TCP
extern	int		read( int, char *, int );
extern	int		close( int );
#endif

/*********************************************************************
                             OpenLink()

	- opens a link to the transputer (i.e the server)

	effects  : returns either
			ER_LINK_BAD     (link already open)
			ER_LINK_CANT    (transputer doesn't respond)
			ER_LINK_SOFT    (link software failure)
		   or a LINK value      (success)
**********************************************************************/

PUBLIC LINK OpenLink ( BYTE *Name )
{
     int	s;
	 int	tmp;
	 int	param_count;
     int	xa;
	 int	xb;
	 int	xc;
	 int	xd;
  struct	sockaddr_in sin;
  struct	sockaddr_in rsin;
  struct	hostent *hp;

	union
	{
		unsigned long l_addr;
		unsigned char addr[4];
	} address;

/**** Get true IP address of host machine ****/

	if (ActiveLink != NULL_LINK)
	{
		return(ER_LINK_BAD);
	}

	if ((*Name == 0) || (Name == NULL))
	{
		strcpy(Name, DefaultDevice);
	}

	hp = gethostbyname(Name);

	s = socket(AF_INET, SOCK_STREAM, 0 );

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl (INADDR_ANY);
	sin.sin_port = htons (0);

	if (bind(s, (struct sockaddr *) & sin, (short int) sizeof (sin)) < 0)
	{
		DEBUG(( "cant bind, err=%d\n", ERRORCODE ));
		return(ER_LINK_CANT);
	}

	if (hp == NULL)
	{
		param_count = sscanf(Name, "%d.%d.%d.%d", &xa, &xb, &xc, &xd);
		address.addr[0] = (unsigned char) xa;             
		address.addr[1] = (unsigned char) xb;
		address.addr[2] = (unsigned char) xc;
		address.addr[3] = (unsigned char) xd;

		if (param_count == 4)
		{
			rsin.sin_addr.s_addr = address.l_addr;
		}
		else
		{
			DEBUG(( "bad internet address\n" ));
			return(ER_LINK_BAD);
		}
	}
	else	/* Name is a textual name */
	{
		memcpy((char*)&rsin.sin_addr, (char*)hp->h_addr, hp->h_length );
	}

	rsin.sin_family = AF_INET;
	rsin.sin_port = htons( HOSTLINK_PORT );

	DEBUG(( "remote  addr is %s, port is %d\n", inet_ntoa(rsin.sin_addr), HOSTLINK_PORT ));

	tmp = connect(s, (struct sockaddr *) & rsin, sizeof (rsin));

	if (tmp < 0)
	{
		DEBUG(( "cant connect, err = %d\n", ERRORCODE ));
		return( ER_LINK_CANT );
	}

	DEBUG(( "Internet connection established with remote %s\n", inet_ntoa( rsin.sin_addr) ));

/**** start sequence numbers as zero ****/

	h_seq = 0;
	t_seq = 0;

/**** send a SYN frame, receiving an ACK indicates success ****/

	DEBUG(( "Sending SYN frame to remote" ));

	if (prot_send (s, rp_SYN, h_seq, 2, TRANS_TIMEOUT) != rp_ACK)
	{
		DEBUG(("Remote Communication link to Transputer not working"));
		close(s);
		h_seq = (h_seq + 1) & 0x007f;
		return( ER_LINK_CANT );
	}
	else
	{
		DEBUG(("Communication GOOD"));

		ActiveLink = s;
		h_seq = (h_seq + 1) & 0x007f;
		return(ActiveLink);
	}
}

/**********************************************************************
                             CloseLink()

	- closes the link between the transputer & the host

	effects  : returns either
			ER_LINK_BAD     (link wasn't open)
			ER_LINK_CANT    (transputer doesn't respond)
			ER_LINK_SOFT    (link software failure)
			SUCCEED         (success)
**********************************************************************/

PUBLIC int CloseLink ( LINK LinkId )
{
	if (LinkId != ActiveLink)
	{
		return (ER_LINK_BAD);
	}
  
/* terminate the connection to the transputer */

	if (prot_send (ActiveLink, rp_RST, h_seq, 2, TRANS_TIMEOUT) != rp_ACK)
	{
		return (Result);
	}
	else
	{
		close(ActiveLink);
		ActiveLink = NULL_LINK;
		return(SUCCEEDED);     /* return success value */
	}
}


/**********************************************************************
                           ReadLink()

	- tries to read Count bytes from the link!

	effects : tries to read rp_MAX_DATA_SIZE bytes from the link,
		  returns either
			ER_LINK_BAD     (link not open)
			ER_LINK_CANT    (transputer doesn't respond)
			ER_LINK_SOFT    (link software failure)
			the number of bytes (up to the amount requested
			 in "Count") successfully read
**********************************************************************/

PUBLIC int ReadLink ( LINK			LinkId,
                      char *		Buffer,
                      unsigned int	Count,
                      int			Timeout )
{
	int got = 0;

	if ((int)LinkId != (int)ActiveLink)
	{
		return(ER_LINK_BAD);
	}

	if (Count < 1)
	{
		return(ER_LINK_CANT);
	}

/**** restrict size of read to maximum ****/

	if (Count > rp_MAX_DATA_SIZE)
	{
		Count = rp_MAX_DATA_SIZE;
	}

/**** no buffered data, so read some more data ****/

	if (saved_len <= 0)
	{
		saved_len = data_read(ActiveLink, &saved_buf, rp_MAX_DATA_SIZE, Count, (short int)Timeout );

		if (saved_len < 0)
		{
			DEBUG (( "ReadLink - returning %d\n", got ));
			return (saved_len);
		}
	}

/**** if read more than wanted, just return wanted ****/

	if (saved_len >= (int)Count)
	{
		got = Count;
	}
	else	/* if read less than wanted, return read */
	{
		got = saved_len;                    
	}

	memcpy (Buffer, saved_buf, got);
	saved_len = saved_len - got;
	saved_buf = saved_buf + got;
	DEBUG(( "ReadLink - returning %d\n", got ));
	return (got);
}


/**********************************************************************
                               WriteLink()

	- tries to write data to the link!

	effects  : tries to write up to "Count" bytes to the link.
		   returns either
			ER_LINK_BAD     (link wasn't open)
			ER_LINK_CANT    (transputer doesn't respond)
			ER_LINK_SOFT    (link software failure)
		   or number of bytes sucessfully sent
**********************************************************************/

PUBLIC int WriteLink ( LINK LinkId,
                       char *Buffer,
                       short int Count,
                       int Timeout )
{
	short int	i;
	short int	rem;
	short int	status;

	rem = Count;

	if (LinkId != ActiveLink)
	{
		return(ER_LINK_BAD);
	}

	if (rem < 1)
	{
		return(ER_LINK_CANT);
	}

	while (rem > 0)
	{
		i = (rem > rp_MAX_DATA_SIZE) ? rp_MAX_DATA_SIZE : rem;

		status = data_send (ActiveLink, rp_DATA, Buffer, i);

		if (!status)
		{
			return(Result);
		}

		Buffer += i;
		rem -= i;
	}

	return (Count);
}

/**********************************************************************
                               ResetLink()

	- reset the transputer's subsystem

	effects  : returns either
			ER_LINK_BAD     (link wasn't open)
			ER_LINK_CANT    (transputer doesn't respond)
			ER_LINK_SOFT    (link software failure)
			SUCCEDDED
***********************************************************************/

PUBLIC int ResetLink ( LINK LinkId )
{
	int code;

	if (LinkId != ActiveLink)
	{
		return (ER_LINK_BAD);
	}

/**** reset the transputer's subsystem ****/

	code = prot_send (ActiveLink, rp_RESET, h_seq, 0, TRANS_TIMEOUT);

	if (code == rp_NACK)
	{
		return (ER_LINK_BAD);
	}

	if (code != rp_ACK)
	{
		return (Result);     
	}
	else
	{
		h_seq = (h_seq + 1) & 0X007f;
		return (SUCCEEDED);     /* return success value */
	}
}

/**********************************************************************
                             AnalyseLink()

	effects  : sets analyse mode on the transputer
		   returns either
			ER_LINK_BAD     (link wasn't open)
			ER_LINK_CANT    (transputer doesn't respond)
			ER_LINK_SOFT    (link software failure)
			SUCCEEDED
**********************************************************************/

PUBLIC int AnalyseLink ( LINK LinkId )
{
	int	code;
  
	if (LinkId != ActiveLink)
	{
		return(ER_LINK_BAD);
	}

	code = prot_send (ActiveLink, rp_ANALYSE, h_seq, 0, TRANS_TIMEOUT);

	if (code == rp_NACK)
	{
		return(ER_LINK_BAD);
	}

	if (code != rp_ACK)
	{
		return(Result);
	}
	else
	{
		h_seq = (h_seq + 1) & 0x007f;
		return(SUCCEEDED);     /* return success value */
	}
}

/**********************************************************************
                               PeekLink()

	- tries to peek 'words' words of data from the transputer

	modifies : Result
	effects  : returns either
			ER_LINK_BAD     (link wasn't open)
			ER_LINK_CANT    (transputer doesn't respond)
			ER_LINK_SOFT    (link software failure)
			0               (no data could be peeked in the available time)
		   or number of bytes successfully peeked
***********************************************************************/

PUBLIC int PeekLink ( LINK LinkId,
                      char *Buffer,
                      u_long address,
                      short int words )
{
	   prot_dgram	resp;
              int	code;
			  int	tries;
			  int	data_ok;
			  int	bytes_returned;                         
	unsigned char	next_t_seq;
	unsigned char	next_h_seq;

	if (LinkId != ActiveLink)
	{
		return (ER_LINK_BAD);
	}

	next_t_seq = (t_seq + 1) &0x007f;
	next_h_seq = (h_seq + 1) &0x007f;

	for (tries = 1; tries < MAX_RETRIES; tries++)
	{
		code = peek_send (ActiveLink, h_seq, address, words);

		if (code == rp_NACK)
		{
			h_seq = next_h_seq;
			return (ER_LINK_BAD);
		}

		if (code != rp_ACK)
		{
			h_seq = next_h_seq;
			return(Result);
		}
		else
		{
			data_ok = timed_read (ActiveLink, (char *) &inudp, sizeof(inudp), timeout_period * tries);

			if (data_ok >= 2 && (inudp.op == rp_DATA && inudp.seq == t_seq))
			{
				h_seq = next_h_seq;
				t_seq = next_t_seq;
				resp.op = rp_ACK;
				resp.seq = inudp.seq;
				resp.len = 0;

				if (send (ActiveLink, (char *) &resp, sizeof(resp), 0) < 0)
				{
					DEBUG(("*fatal error : PeekLink[1] - send failed=%d\n", ERRORCODE ));
					return(ER_LINK_SOFT); 
				}

				bytes_returned = ntohs(inudp.len);
				DEBUG(( "PeekLink :returning %d\n",bytes_returned ));
				memcpy(Buffer, inudp.data, bytes_returned);

				return(bytes_returned);
			}
			else
			{                                                          
				if (data_ok >= 2 && inudp.op != rp_ACK && inudp.op != rp_NACK &&
				    inudp.op != rp_CONT && inudp.op != rp_ERROR )
				{
					resp.op = rp_ACK;
					resp.seq = inudp.seq;
					resp.len = 0;

					if (send (ActiveLink, (char *) &resp, sizeof (resp), 0) < 0)
					{  
						DEBUG(("*fatal error : PeekLink[2] - send failed=%d\n", ERRORCODE ));
						return(ER_LINK_SOFT);
					}
				}
			}
		}
	}

	t_seq = next_t_seq;
	h_seq = next_h_seq;

	return(ER_LINK_CANT);
}

/**********************************************************************
                           TestError()

	- tests the error state of the transputer

	modifies : h_seq (incremented)
	effects  : tests the error state of the transputer
		   returns either
			ER_LINK_BAD     (link wasn't open)
			ER_LINK_CANT    (transputer doesn't respond)
			ER_LINK_SOFT    (link software failure)
			0               (error flag not set)
			1               (error flag set)
***********************************************************************/

PUBLIC int TestError ( LINK LinkId )
{
	  prot_dgram	reply;
	 error_dgram	resp;
	 error_dgram	error_inq;
	         int	tries;
		     int	reply_ok;

	if (LinkId != ActiveLink)
	{
		return (ER_LINK_BAD);
	}
  
	error_inq.op = rp_ERROR;
	error_inq.seq = h_seq;

	for (tries = 1; tries < MAX_RETRIES; tries++)
	{
		if (send (ActiveLink, (char *) &error_inq, sizeof (error_inq), 0) < 0)
		{
			DEBUG (("*fatal error : TestError[1] - send failed =%d\n", ERRORCODE ));
			return (ER_LINK_SOFT);
		}
		else
		{
			reply_ok = (timed_read (ActiveLink, (char *) &resp, sizeof (resp), timeout_period * tries) > 0);

			if (reply_ok > 0)
			{
				if (resp.op == rp_NACK)
				{
					return (ER_LINK_BAD);
				}

/**** reply obtained, so return error status of transputer ****/

				if (resp.op == rp_ERROR)
				{
					h_seq = (h_seq + 1) & 0x007f;
					return (int) (resp.status);
				}
				else
				{
					if (resp.op != rp_ACK && resp.op != rp_NACK && resp.op != rp_CONT && resp.op != rp_ERROR)
					{
						reply.op = rp_ACK;
						reply.seq = resp.seq;
						reply.len = htons ((short int) 0);
						reply.timeout = htons ((short int) 0);

						if (send (ActiveLink, (char *) &reply, sizeof (reply), 0) < 0)
						{
							DEBUG (("*fatal error : TestError[2] - send failed =%d\n", ERRORCODE ));
							return (ER_LINK_SOFT); 
						}
					}
				}
			}
		}
	}

  /* failed */

  h_seq = (h_seq + 1) &0x007f;
  return (ER_LINK_CANT);
};


/**********************************************************************
                               TestRead()

	- test the read operation
	effects : returns either
			ER_LINK_BAD     (link wasn't open)
			0               (link might not be able to read)
***********************************************************************/

PUBLIC int TestRead ( LINK LinkId )
{
	if (LinkId != ActiveLink)
	{
		return (ER_LINK_BAD);
	}
	else
	{
		return (0);
	}
}

/**********************************************************************
                              TestWrite()

	- test the write operation
	effects : returns either
			ER_LINK_BAD     (link wasn't open)
			0               (link might not be able to write)
***********************************************************************/

PUBLIC int TestWrite ( LINK LinkId )
{
	if (LinkId != ActiveLink)
	{
		return (ER_LINK_BAD);
	}
	else
	{
		return (0);
	}
}

/**********************************************************************
                               TestLink()

	- tests the link is operational

	effects  : tests the link status. returns either
			ER_LINK_BAD     (link wasn't open)
			ER_LINK_CANT    (link communication failed)
			SUCCEEDED       (link is ok)
***********************************************************************/

PUBLIC TestLink ( LINK LinkId )
{
	int status;

	if (LinkId != ActiveLink)
	{
		return (ER_LINK_BAD);
	}

	status = prot_send (ActiveLink, rp_ACTIVE, h_seq, 0, TRANS_TIMEOUT);

	if (status == rp_NACK)
	{
		return (ER_LINK_BAD);
	}

	if (status == rp_TIMEOUT)
	{
		return (Result);
	}
	else
	{
		h_seq = (h_seq + 1) &0x007f;
		return (SUCCEEDED);
	}
}

/************************ INTERNAL FUNCTIONS ******************************/

/* timeout functions */

/**********************************************************************
***********************************************************************/

PRIVATE int timed_read ( int      s,
                         char *   buf,
                         int      len,
                         long int millisec)
{
	   int fdwidth;
	fd_set readfds;
	struct timeval timeout;

	assert( s < FD_SETSIZE );

	timeout.tv_sec = millisec / 1000;
	timeout.tv_usec = (millisec % 1000) * 1000;

	fdwidth = s + 1;

	FD_ZERO( &readfds );
	FD_SET( s, &readfds );

	if ( select(fdwidth, &readfds, (fd_set *)NULL, (fd_set *)NULL, &timeout) > 0 )
	{
		return( read (s, buf, len) );
	}
	else
	{
		return( 0 );
	}
}

/**********************************************************************
                           prot_send()

	- send a protocol frame to the transputer & obtain a reply

	modifies : Result
	effects  : sends a frame to the transputer and returns either
		   ACK, NACK or TIMEOUT (TIMEOUT occurs if an ACK or
		   NACK of the correct sequence number is received
		   within the timeout period. the timeout parameter
		   passed is sent to the transputer & used for its own
		   timeout checking. If TIMEOUT is returned, inspect
		   Result for status value.
***********************************************************************/

PRIVATE unsigned char prot_send ( LINK          s,
                                  unsigned char op,
                                  unsigned char seq,
                                  short int     len,
                                  short int     timeout )
{
  prot_dgram	requ;
  prot_dgram	resp;
         int	tries;
		 int	reply_ok;
		 int	numbytes;
		 int	timeval1;
		 int	timeval2;

	requ.op = op;
	requ.seq = seq;
	requ.len = htons ((short int) len);
	requ.timeout = htons ((short int) timeout);

/*
* make upto MAX_RETRIES tries to successfully send a frame to the
* transputer, and get either an ACK or NACK with the correct sequence
* number back
*/

	DEBUG(("(Sending protocol packet to remote)\n" ));

	Result = SUCCEEDED;

	for (tries = 1; tries < MAX_RETRIES; tries++)
	{
/* send request packet */

		if (send (s, (char *) &requ, sizeof (requ), 0) < 0)
		{
			DEBUG (( "*fatal error : prot_send - send failed =%d\n", ERRORCODE ));
			Result = ER_LINK_SOFT;
			return (rp_TIMEOUT);
		}
		else
		{
			DEBUG(("(Sent protocol to remote Transputer)\n"));

/* request sent ok, so get reply */

			time ((time_t*)&timeval1);
			reply_ok = FALSE;

			while (!reply_ok)
			{
				numbytes = timed_read( s, (char *)&resp, sizeof(resp), timeout_period * tries );

				if (numbytes > 0) /* got a reply */
				{
					reply_ok = TRUE;

					if  ((resp.op == rp_ACK || resp.op == rp_NACK) && (resp.seq == seq))
					{
						return (resp.op);             
					}

				}

				time ((time_t*)&timeval2);

				if ((timeval2-timeval1) >= QUIETTIMEOUT) 
				{
					reply_ok = TRUE;
				}
			}
		}
	}

/* failed */

	DEBUG(("(Reply timed out waiting)\n"));
	Result = ER_LINK_CANT;
	return(rp_TIMEOUT);
}

/**********************************************************************
                              peek_send()

	- send a peek frame to the transputer & obtain a reply

	modifies : Result
	effects  : sends a frame to the transputer and returns either
		   ACK, NACK or TIMEOUT (TIMEOUT occurs if an ACK or
		   NACK of the correct sequence number is received
		   within the timeout period. the timeout parameter
		   passed is sent to the transputer & used for its own
		   timeout checking. if TIMEOUT is returned, inspect
		   Result for status value.
**********************************************************************/

PRIVATE unsigned char peek_send ( LINK          s,
                                  unsigned char seq,
                                  unsigned long address,
                                  short int     numwords )
{
  peek_dgram	requ;
  prot_dgram	resp;
         int	tries;
		 int	reply_ok;
		 int	numbytes;
		 int	timeval1;
		 int	timeval2;

	requ.op = rp_PEEK;
	requ.seq = h_seq;
	requ.address = htonl(address);
	requ.numwords = htons(numwords);

/*
* make upto MAX_RETRIES tries to successfully send a frame to the
* transputer, and get either an ACK or NACK with the correct sequence
* number back
*/
	Result = SUCCEEDED;

	for (tries = 1; tries < MAX_RETRIES; tries++)
	{
/**** send request packet ****/

		if (send (s, (char *) &requ, sizeof (requ), 0) < 0)
		{
			DEBUG (("*fatal error : peek_send - send failed =%d\n", ERRORCODE ));
			Result == ER_LINK_SOFT;
			return (rp_TIMEOUT);
		}
		else
		{
/**** request sent ok, so get reply ****/

		reply_ok = FALSE;
		time ((time_t*)&timeval1);

			while (!reply_ok)
			{
				numbytes = timed_read (s, (char *) &resp, sizeof (resp), timeout_period * tries);

				if (numbytes > 0)
				{
					reply_ok = TRUE;

					if ((resp.op == rp_ACK || resp.op == rp_NACK) && (resp.seq == seq))
					{
						return (resp.op);
					}
				}

/**** if X seconds elapsed cant contact transputer ****/

				time ((time_t*)&timeval2);

				if ((timeval2-timeval1) >= QUIETTIMEOUT)
				{
					reply_ok = TRUE;
				}
			}
		}
	}

/**** failed ****/

  Result = ER_LINK_CANT;
  return(rp_TIMEOUT);
}

/**********************************************************************
                           data_send()

	- send data to the transputer & obtain a reply
	modifies : Result
	effects   : returns either
			TRUE    (data send successfully)
			FALSE   (send failed -> status returned in Result)
***********************************************************************/

PRIVATE int data_send ( LINK          s,
                        unsigned char op,
                        char *        buff,
                        short int     len )
{
  data_dgram	requ;
  prot_dgram	resp;
  prot_dgram	reply;
          int	tries;
		  int	rcount;
		  int	reply_ok;
		  int	tresult;
		  int	timed_out;

	Result = SUCCEEDED;

	memcpy (requ.data, buff, len);
	requ.op = op;
	requ.seq = h_seq;
	requ.len = htons ((short int) len);

	DEBUG (( "data_send: sending %d\n", len ));

	len += rp_DATA_HEADER;

	for (tries = 1; tries < MAX_RETRIES; tries++)
	{
/**** send request packet ****/

		if (send (s, (char *) &requ, sizeof (requ), 0) < 0)
		{
			Result = ER_LINK_SOFT;
			return (FALSE);
		}
		else
		{
			reply_ok = FALSE;
			timed_out = FALSE;

			while (!timed_out && !reply_ok)
			{
				tresult = timed_read (s, (char *) &resp, sizeof (resp), timeout_period * tries);
				timed_out = (tresult == 0);
				reply_ok = (!timed_out && (resp.op == rp_ACK || resp.op == rp_NACK || 
				resp.op == rp_CONT || resp.op == rp_SUSP) && resp.seq == h_seq);

				if (reply_ok && resp.op == rp_NACK)
				{
					Result = ER_LINK_BAD;
					return (Result);
				}

				if (!timed_out && resp.op == rp_DATA)
				{
					DEBUG (( "received DATA frame in write\n"));
					reply.op = rp_ACK;
					reply.seq = resp.seq;
					reply.len = 0;

					if (send (s, (char *) &reply, sizeof (reply), 0) < 0)
					{
						DEBUG (( "*fatal error : data_send[2] - send failed =%d\n", ERRORCODE ));
						Result = ER_LINK_SOFT;
						return (FALSE);
					}
				}
			}

/**************************************/

			if (reply_ok)
			{
				if ((resp.op == rp_SUSP) && (resp.seq == h_seq))
				{
					reply_ok = FALSE;
					timed_out = FALSE;

					while (!timed_out && !reply_ok)
					{
						rcount = timed_read (s, (char *) &resp, sizeof (resp),
						timeout_period * tries);
						timed_out = (rcount == 0);

						if (!timed_out && resp.seq == h_seq && resp.op == rp_ACK)
						{
							reply_ok = TRUE;
							h_seq = (h_seq + 1) &0x007f;
							return (TRUE);
						}
					}

					tries = 1;    /* transputer didn't do request so retry*/
				}
				else
				{
					if (resp.op == rp_ACK)
					{
						h_seq = (h_seq + 1) & 0x007f;
						return (TRUE);
					}
					else
					{
						DEBUG (( "Invalid sp frame received %d -- Please report\n", resp.op ));
						h_seq = (h_seq + 1) & 0x007f;
						Result = ER_LINK_CANT;
						return (FALSE);
					}
				}
			}
		}
	}

/* failed */

  h_seq = (h_seq + 1) & 0x007f;
  Result = ER_LINK_CANT;
  return (FALSE);
}         

/**********************************************************************
                              data_read()

	- read data from transputer

	effects  : returns either
		    ER_LINK_BAD     (link wasn't open)
		    ER_LINK_CANT    (link communication failed)
		    ER_LINK_SOFT    (link software failure)
		    0               (no data could be read in available time)
		   or number of bytes successfully read
***********************************************************************/

PRIVATE int data_read ( LINK      s,
                        char **   buff,
                        short int maxlen,
                        short int len,
                        short int timeout )
{
     prot_dgram	resp;
            int	tries, data_ok, code;
  unsigned char	next_t_seq, next_h_seq;

	next_t_seq = (t_seq + 1) & 0x007f;
	next_h_seq = (h_seq + 1) & 0x007f;

	Result = SUCCEEDED;

	for (tries = 1; tries < MAX_RETRIES; tries++)
	{
		code = prot_send (s, rp_REQ, h_seq, len, timeout);

		if (code == rp_NACK)
		{
			h_seq = next_h_seq;
			return (ER_LINK_BAD);
		}

		if (code == rp_TIMEOUT)
		{
			h_seq = next_h_seq;
			return (Result);
		}
		else
		{
			data_ok = timed_read (s, (char *) &inudp, sizeof (inudp), timeout_period * tries);

			if (data_ok >= 2 && inudp.op == rp_DATA && inudp.seq == t_seq)
			{
				h_seq = next_h_seq;
				t_seq = next_t_seq;
				resp.op = rp_ACK;
				resp.seq = inudp.seq;
				resp.len = 0;

				if (send (s, (char *) &resp, sizeof (resp), 0) < 0)
				{
					DEBUG (( "*fatal error : data_read[1] - send failed =%d\n", ERRORCODE ));
					return (ER_LINK_SOFT);
				}

				DEBUG (( "input len = %d\n", len ));
				*buff = (char *) &(inudp.data[0]);
				DEBUG (( "data_read: returning %d\n", ntohs (inudp.len) ));
				return (ntohs (inudp.len));
			}
			else
			{
				if ((data_ok >=2) && (inudp.op != rp_ACK) && (inudp.op != rp_NACK) &&
				    (inudp.op != rp_CONT) && (inudp.op != rp_ERROR))
				{
					resp.op = rp_ACK;
					resp.seq = inudp.seq;
					resp.len = 0;

					if (send (s, (char *) &resp, sizeof (resp), 0) < 0)
					{
						DEBUG (( "*fatal error : data_read[2] - send failed =%d\n", ERRORCODE ));
						return (ER_LINK_SOFT);
					}
				}
			}
		}
	}

  return (ER_LINK_CANT);           
}

/*******************************  EOF  *******************************/


